package Bathroom;

use POE;

sub new {
	my $class = shift;

	my $self = bless {
		room_name => 'Bathroom',
	},$class;

	my $session = POE::Session->create(
		#options => { default=>1, trace=>1 },
		heap => shift,
		object_states => [
			$self => {
				_start			=> '_start',
				_stop			=> '_stop',
				shutdown		=> 'shutdown',
				
				# provided for us
				walked_in		=> 'walked_in',
				walked_out		=> 'walked_out',
			}
		],
	);

	$self->{sid} = $session->ID;

	return $self;
}


sub _start {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	$kernel->call($heap->{m} => _write_log => { v => 4, msg => "$self->{room_name} session started" });

	# keep us alive
    $kernel->refcount_increment($session->ID, "rooms::".__PACKAGE__);
}

sub _stop {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	
}

sub shutdown {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	$kernel->alarm_remove_all();
    $kernel->refcount_decrement($session->ID, "rooms::".__PACKAGE__);

	return $session->ID;
}

sub walked_out {
	my ($kernel, $heap, $session, $sender, $self, $sid, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0, ARG1];

	delete $heap->{uids}{$o->{uid}};
	
	my $txt;	
	if ($o->{direction}) {
		my %op = (
			n => 'to the north',
			s => 'to the south',
			e => 'to the east',
			w => 'to the west',
			sw => 'to the southwest',
			nw => 'to the northwest',
			ne => 'to the northeast',
			se => 'to the southeast',
			d => 'down',
			u => 'up',
		);
		$txt = sprintf("%s went %s",$heap->{data}->{users}{$o->{uid}}->{alias},$op{$o->{direction}});
	} else {
		$txt = sprintf("\0[bold;white]*poof*\0[reset] %s disappeared!",$heap->{data}->{users}{$o->{uid}}->{alias});
	}
	foreach my $u (keys %{$heap->{uids}}) {
		next if ($u == $o->{uid});
		$kernel->post($heap->{uids}{$u} => send => $txt);
	}
}

sub walked_in {
	my ($kernel, $heap, $session, $sender, $self, $sid, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0, ARG1];

	# whos here..
	$heap->{uids}{$o->{uid}} = $sid;

	$kernel->post($sid => send => "\0[bold;white]Bathroom\0[reset] This is the bathroom.",
		"There is a toilet here; the lid is up");
	my $txt;
	if ($o->{direction}) {
		my %op = (
			n => 'from the north',
			s => 'from the south',
			e => 'from the east',
			w => 'from the west',
			sw => 'from the southwest',
			nw => 'from the northwest',
			ne => 'from the northeast',
			se => 'from the southeast',
			d => 'from inside toilet. How? You are unsure.', # why connect a down entrance in a bathroom!? haha
			u => 'from a panel in the ceiling',
		);
		$txt = sprintf("%s arrived %s",$heap->{data}->{users}{$o->{uid}}->{alias},$op{$o->{direction}});
	} else {
		$txt = sprintf("\0[bold;white]*poof*\0[reset] %s appeared out of nowhere!",$heap->{data}->{users}{$o->{uid}}->{alias});
	}
	foreach my $u (keys %{$heap->{uids}}) {
		next if ($u == $o->{uid});
		$kernel->post($heap->{uids}{$u} => send => $txt);
	}
}

1;

