package Network;
use strict;
use vars qw( $VERSION @ISA );

require Exporter;
@ISA = qw( Exporter );
$VERSION = '0.01';

use IO::Socket;
use IO::Select;
use Net::Telnet qw(
    TELNET_IAC
    TELNET_DONT
    TELNET_DO
    TELNET_WONT
    TELNET_WILL
    TELNET_GA
    TELNET_EOR

    TELNET_EC
    TELNET_EL
    TELNET_IP
    TELNET_AYT

    TELOPT_ECHO
    TELOPT_SGA
    TELOPT_EOR
);

use Fionread;

use Flags;
use Misc;

use constant CR   => "\015";
use constant LF   => "\012";
use constant CRLF => CR . LF;

my $Listen;   # listen for new connections
my $Who;      # who port
my $Exam;     # examine port
my $Mail;     # mail port
my $Select;   # look for activity on open sockets

sub plug_in {
    my($port,$who,$exam,$mail) = @_;

    $Listen = IO::Socket::INET->new(
        Proto => 'tcp',
        LocalPort => $port,
        Listen => 5,
        Reuse => 1,
    );
    unless (defined $Listen) {
        &slog("Failed to make listen socket: $!\n");
        return;
    }

    $Who = IO::Socket::INET->new(
        Proto => 'tcp',
        LocalPort => $who,
        Listen => 5,
        Reuse => 1,
    );
    unless (defined $Who) {
        &slog("Failed to make who port: $!\n");
        return;
    }

    $Exam = IO::Socket::INET->new(
        Proto     => 'tcp',
        LocalPort => $exam,
        Listen    => 5,
        Reuse     => 1,
    );
    unless (defined $Exam) {
        &slog("Failed to make examine port: $!\n");
        return;
    }

    $Mail = IO::Socket::INET->new(
        Proto     => 'tcp',
        LocalPort => $mail,
        Listen    => 5,
        Reuse     => 1,
    );
    unless (defined $Mail) {
        &slog("Failed to make mail port: $!\n");
        return;
    }

    $Select = IO::Select->new($Listen, $Who, $Exam, $Mail);

    &slog("Listening on port $port...\n");
    &slog("Who port is $who...\n");
    &slog("Exam port is $exam...\n");

    1;
}

sub scan_sockets {
    my @ready;

    @ready = $Select->can_read($Conf::Sync_Time);

    foreach my $socket (@ready) {
        if ($socket == $Listen) {
            &accept_connection();
        }
        elsif ($socket == $Who) {
            &send_who();
        }
        elsif ($socket == $Exam) {
            &accept_rexam();
        }
        elsif ($socket == $Mail) {
            &accept_mail();
        }
        else {
            my $fd = $socket->fileno;
            my $user = $User::fd2user{$fd} || '';

            unless ($user) {
                &slog("Can't map fd `$fd' to user!\n");
                $Select->remove($socket);
            }
            elsif (not ref $user) {
                &slog("Bad entry in fd -> user map\n");
                $Select->remove($socket);
            }
            else {
                &get_input($user);
            }
        }
    }
}

sub accept_connection {
    my $new = $Listen->accept;

    unless (defined $new) {
        &slog("Failed accept: $!\n");
        return;
    }
    $Select->add($new);

    my $u = &User::new($new);

    &User::tell($u, sprintf("%c%c%c", TELNET_IAC,
                                      TELNET_WILL,
                                      TELOPT_EOR));

    $Serv::Num_Users++;
    &User::connect($u);
}

sub accept_rexam {
    my $new = $Exam->accept;

    unless (defined $new) {
        &slog("Failed accept: $!\n");
        return;
    }
    $Select->add($new);

    my $u = &User::new($new);
    $u->{'rows'} = 0;
    $u->{'flags'} |= REXAM;

    $Serv::Num_Exam++;
    &User::rexam($u);
}

sub accept_mail {
    my $new = $Mail->accept;

    unless (defined $new) {
        &slog("Failed accept: $!\n");
        return;
    }
    $Select->add($new);

    my $u = &User::new($new);
    $u->{'rows'} = 0;
    $u->{'flags'} |= RMAIL;

    $Serv::Num_Mail++;
    &User::rmail($u);
}

sub send_who {
    my $new = $Who->accept;

    unless (defined $new) {
        &slog("Failed accept: $!\n");
        return;
    }

    my $u = &User::new($new);
    $u->{'rows'} = 0;
    $u->{'flags'} |= RWHO;

    &Command::who($u, '');

    &User::quit($u);
}

sub get_input {
    my $u = shift;
    my $s = $u->{'fd'};

    my $len;
    $len = Fionread::chars_ready($s);

    unless (defined $len) {
        &slog("Failed FIONREAD ioctl: $!\n");
        &User::quit($u);
        return;
    }

    if ($len == 0) {
       &User::quit($u);
       return;
    }

    my $ch;
    for (; $len > 0; $len--) {
        unless (defined $s->recv($ch, 1)) {
            &slog("Failed read from $u->{'IP'}: $!\n");
            &User::quit($u);
            return;
        }
        else {
            my $i = ord $ch;
            if ($i == 255) {
                $u->{'flags'} &= ~(LAST_WAS_R | LAST_WAS_N);
                &handle_telnet($u);
                return;
            }
            elsif ($ch eq LF) {
                if (not ($u->{'flags'} & LAST_WAS_R)) {
                    $u->{'flags'} |= (LAST_WAS_N | INPUT_READY);
                    $u->{'col'} = 0;
                }
                return;
            }
            elsif ($ch eq CR) {
                if (not ($u->{'flags'} & LAST_WAS_N)) {
                    $u->{'flags'} |= (LAST_WAS_R | INPUT_READY);
                    $u->{'col'} = 0;
                }
                return;
            }
            elsif ($i == 247 or $i == 8 or $i == 127) {
                backspace($u);
                #return;
            }
            else {
                $u->{'flags'} &= ~(LAST_WAS_R | LAST_WAS_N);

                if ($i > 31) {
                    $u->{'input'} .= $ch;

                    if (not ($u->{'flags'} & PASSWORD_MODE)
                        and $u->{'flags'} & DO_LOCAL_ECHO)
                    {
                        unless ($s->send($ch)) {
                            my $ip = $u->{'IP'} || '<unknown IP>';
                            &slog("Echoing back to user from $ip\n");
                            &User::quit($u);
                            return;
                        }
                    }
                }
                #return;
            }
        }
    }
}

sub handle_telnet {
    my $u = shift;
    my $ch;

    my $s = $u->{'fd'};
    unless ($s->recv($ch, 1)) {
        return;
    }

    my $i = ord($ch);
    if ($i == TELNET_EC) {
        &backspace($u);
    }
    elsif ($i == TELNET_EL) {
        $u->{'input'} = '';
    }
    elsif ($i == TELNET_IP) {
        &User::quit($u);
    }
    elsif ($i == TELNET_AYT) {
        $s->send("[ Yes ]" . CRLF);
    }
    elsif ($i == TELNET_DO) {
        unless ($s->recv($ch, 1)) {
            return;
        }

        $i = ord($ch);
        if ($i == TELOPT_ECHO) {
            unless ($u->{'flags'} & DO_LOCAL_ECHO) {
                $u->{'flags'} |= DO_LOCAL_ECHO;
                $s->send(sprintf("%c%c%c", TELNET_IAC,
                                           TELNET_WILL,
                                           TELOPT_ECHO));
            }
        }
        elsif ($i == TELOPT_SGA) {
            unless ($u->{'flags'} & DO_SGA) {
                $u->{'flags'} |= DO_SGA;
                $s->send(sprintf("%c%c%c", TELNET_IAC,
                                           TELNET_WILL,
                                           TELOPT_SGA));
            }
        }
        elsif ($i == TELOPT_EOR) {
            unless ($u->{'flags'} & DO_EOR) {
                $u->{'flags'} |= DO_EOR;
                $s->send(sprintf("%c%c%c", TELNET_IAC,
                                           TELNET_WILL,
                                           TELOPT_EOR));
            }
        }
    }
    elsif ($i == TELNET_DONT) {
        unless ($s->recv($ch, 1)) {
            return;
        }

        $i = ord($ch);
        if ($i == TELOPT_ECHO) {
            if ($u->{'flags'} & DO_LOCAL_ECHO) {
                $u->{'flags'} &= ~(DO_LOCAL_ECHO);
                $s->send(sprintf("%c%c%c", TELNET_IAC,
                                           TELNET_WONT,
                                           TELOPT_ECHO));
            }
        }
        elsif ($i == TELOPT_SGA) {
            if ($u->{'flags'} & DO_SGA) {
                $u->{'flags'} &= ~(DO_SGA);
                $s->send(sprintf("%c%c%c", TELNET_IAC,
                                           TELNET_WONT,
                                           TELOPT_SGA));
            }
        }
        elsif ($i == TELOPT_EOR) {
            if ($u->{'flags'} & DO_EOR) {
                $u->{'flags'} &= ~(DO_EOR);
                $s->send(sprintf("%c%c%c", TELNET_IAC,
                                           TELNET_WONT,
                                           TELOPT_EOR));
            }
        }
    }
}

sub backspace {
    my $u = shift;

    $u->{'input'} =~ s/.$//;
}

sub prompt_user {
    my $u = shift;
    my $msg = shift;

    if ($u->{'flags'} & DO_EOR) {
        &tell_user($u, sprintf("%s%c%c", $msg, TELNET_IAC, TELNET_EOR));
    }
    elsif (not ($u->{'flags'} & DO_SGA)) {
        &tell_user($u, sprintf("%s%c%c", $msg, TELNET_IAC, TELNET_GA));
    }
}

sub tell_user {
    my $u = shift;
    my $msg = shift;

    return unless $msg;

    if (not defined $u->{'fd'} or $u->{'flags'} & DUMP_ME) {
        return;
    }

    return if $u->{'flags'} & QUIET;

    if ($Serv::Flags & MONITORED and $u->{'sflags'} & MONITOR) {
        $msg = "<$Serv::Current_User->{'name'}> $msg";
    }

    if (ref $u->{'elog'} and $u->{'elog'}->opened) {
        print { $u->{'elog'} } $msg;
    }

    my $out;
    if ($u->{'wrap'} and length($msg) > ($u->{'cols'} || 79)) {
        $out = &process_output($u, $msg);
    }
    else {
        $out = $msg;
        $out =~ s/\n/${\CRLF}/g;
    }

    if ($Serv::Flags & HILIGHTED and $u->{'sflags'} & HILIGHT) {
        $out = "\e[1m$out\e[0m";
    }

    unless ($u->{'fd'}->opened) {
        &User::quit($u);
        return;
    }

    unless ($u->{'fd'}->send($out)) {
        &User::quit($u);
    }
}

sub process_output {  # XXXXX
    my $u = shift;
    my $str = shift;

    my $i;
    my $idx;
    my $ret = '';

    local($_);
    for ($idx = 0; $idx < length($str);) {
        if (substr($str, $idx, 1) eq LF) {
            $ret .= CRLF;
            $u->{'col'} = 0;
            $idx++;
        }
        else {
            if ($u->{'cols'} and $u->{'col'} >= $u->{'cols'}) {
                for ($i = 0; $i < $u->{'wrap'}; $i++) {
                   last if substr($str, --$idx, 1) =~ /\s/;

                   $ret =~ s/.$//s;
                }

                if ($i != $u->{'wrap'}) {
                    $ret .= CRLF . "  ";
                    $u->{'col'} = 3;
                    $idx++;
                }
                else {
                    $idx += $i;
                    $ret .= CRLF . "  ";
                    $u->{'col'} = 3;
                }
            }

            $ret .= substr($str, $idx++, 1);
            $u->{'col'}++;
        }
    }

    $ret;
}

sub kill_socket {
    my $s = shift;

    $Select->remove($s);
    $s->close;
}

sub no_echo {
    my $u = shift;

    unless ($u->{'flags'} & DO_LOCAL_ECHO) {
        &User::tell($u, sprintf("%c%c%c", TELNET_IAC,
                                          TELNET_WILL,
                                          TELOPT_ECHO));
    }
    $u->{'flags'} |= PASSWORD_MODE;
}

sub do_echo {
    my $u = shift;

    unless ($u->{'flags'} & DO_LOCAL_ECHO) {
        &User::tell($u, sprintf("%c%c%c", TELNET_IAC,
                                          TELNET_WONT,
                                          TELOPT_ECHO));
    }
    $u->{'flags'} &= ~(PASSWORD_MODE);
}

1;
