#!/usr/bin/perl
# $Id$
#
# Copyright 2003 by Rocco Caputo.  Free software.  Modifiable and
# distributable under the same terms as Perl itself.  Have fun!
#
# The first playable release was coded in five hours, between
# 2003-11-03 @23:00 and 2003-11-04 @04:00.
#
# Requires POE 0.27 or newer.  TODO stuff at the end of the file.
#
# You may reach the author at <rcaputo@pobox.com>.
#
# Color stuff added by David Davis  xantus@cpan.org>

use warnings;
use strict;

use lib qw(/projects/lib);
use DB_File;
use YAML qw(freeze thaw);

our %colors = (
  'reset'      => 0,
  'bold'       => 1,
  'dark'       => 2,
  'underline'  => 4,
  'underscore' => 4,
  'blink'      => 5,
  'reverse'    => 7,
  'concealed'  => 8,
  
  'black'      => 30,
  'on_black'   => 40,
  'red'        => 31,
  'on_red'     => 41,
  'green'      => 32,
  'on_green'   => 42,
  'yellow'     => 33,
  'on_yellow'  => 43,
  'blue'       => 34,
  'on_blue'    => 44,
  'magenta'    => 35,
  'on_magenta' => 45,
  'cyan'       => 36,
  'on_cyan'    => 46,
  'white'      => 37,
  'on_white'   => 47,
);

#------------------------------------------------------------------------------
# POE BASED SOCKET SERVER ACTS AS THE USER INTERFACE
#------------------------------------------------------------------------------

use POE qw( Component::Server::TCP Filter::Telnet );

POE::Component::Server::TCP->new(
	Port					=> 23,
	ClientFilter		=> "POE::Filter::Telnet",
	ClientInput				=> \&got_client_input,
	ClientConnected			=> \&got_client_connect,
	ClientError				=> \&got_client_error,
	ClientDisconnected		=> \&got_client_disconnect,
	InlineStates			=> {
		send_message	=> \&send_message,
		welcome_banner	=> \&send_welcome,
		begin_limbo		=> \&begin_limbo,
		remind_limbo	=> \&reminder_limbo,
		cease_limbo		=> \&cease_limbo,
		disconnect		=> \&do_shutdown,
	},
);

map_setup();

# Run bitch, ruuuun! - Shorty
$poe_kernel->run();
exit 0;

### Handle POE events.

sub got_client_connect {
  $_[HEAP]->{colorize} = 0;
  $_[HEAP]->{input_mode} = 'line';
  $_[HEAP]->{input} = '';
  $_[KERNEL]->yield("welcome_banner");
}

sub got_client_disconnect {
  if ($_[HEAP]->{is_in_limbo}) {
    $_[KERNEL]->yield("disconnect");
    return;
  }
  player_disappears($_[SESSION], "\0[bold;white]fa\0[white]de\0[reset]s away.");
}

sub got_client_error {
  my ($kernel, $session, $syscall, $errno) = @_[KERNEL, SESSION, ARG0, ARG1];
  return if $syscall eq "read" and $errno == 0;
  if ($_[HEAP]->{is_in_limbo}) {
    $kernel->yield("disconnect");
    return;
  }
  player_disappears($session, "falls to the floor and \0[bold;white]fa\0[white]de\0[reset]s away.");
}

sub send_welcome {
  personal_effect(
    $_[SESSION]->ID,
    "Please sign up or log in, or learn more about multiwumpus:",
    "  login  (nickname) (password)",
    "  signup (nickname) (password)",
    "  instructions (or info) - For information about the game.",
  );
}

sub got_client_input {
  my ($kernel, $session, $heap, $input) = @_[KERNEL, SESSION, HEAP, ARG0];

  ### Cleanse input.  This happens in all modes.

  # Received telnet command
  if ($input =~ s/\xFF(.*)//) {
	unless ($heap->{telnet}) {
  		$heap->{colorize} = 1;
		personal_effect(
		      $session->ID,
		      "Telnet client detected, \0[green]color\0[reset] is now on. Type color off\nafter logging in if you don't like colors."
		    );
		$heap->{telnet} = 1;
		$heap->{telnet_sga} = 0;
		$heap->{telnet_echo} = 0;
		$heap->{telnet_eor} = 0;
		# IAC DO ECHO
		$kernel->yield(send_message => sprintf("%c%c%c",255,251,1));
	}
	my @cmd = split(/:/,$1);
	# telnet command received
	if ($cmd[0] eq 'TELNET_DO') {
		if ($cmd[1] eq 'TELOPT_SGA') {
			# IAC WILL SGA
			# Suppress Go Ahead
			$kernel->yield(send_message => sprintf("%c%c%c",255,251,3));
			$heap->{telnet_sga} = 1;
		} elsif ($cmd[1] eq 'TELOPT_ECHO') {
			# IAC WILL ECHO
			# Echo
			$kernel->yield(send_message => sprintf("%c%c%c",255,251,1));
			$heap->{telnet_echo} = 1;
			$heap->{input_mode} = 'char';
		} elsif ($cmd[1] eq 'TELOPT_EOR') {
			# IAC WILL EOR
			# End of record (transparent mode)
			$kernel->yield(send_message => sprintf("%c%c%c",255,251,25));
			$heap->{telnet_eor} = 1;
		}
	} elsif ($cmd[0] eq 'TELNET_DONT') {
		if ($cmd[1] eq 'TELOPT_SGA') {
			# IAC WONT SGA
			# Suppress Go Ahead
			$kernel->yield(send_message => sprintf("%c%c%c",255,252,3));
			$heap->{telnet_sga} = 0;
		} elsif ($cmd[1] eq 'TELOPT_ECHO') {
			# IAC WONT ECHO
			# Echo
			$kernel->yield(send_message => sprintf("%c%c%c",255,252,1));
			$heap->{telnet_echo} = 0;
		} elsif ($cmd[1] eq 'TELOPT_EOR') {
			# IAC WONT EOR
			# End of record (transparent mode)
			$kernel->yield(send_message => sprintf("%c%c%c",255,252,25));
			$heap->{telnet_eor} = 0;
		}
	} elsif ($cmd[0] eq 'TELNET_AYT') {
		$kernel->yield(send_message => "[ Yes, I'm here ]");
	} elsif ($cmd[0] eq 'TELNET_IP') {
		# quit
    	player_quit($session, "Telnet client disconnected");
		return;
	} elsif ($cmd[0] eq 'TELNET_EL') {
		# erase line
		print "erase line\n";
		$heap->{input} = '';
		$input = '';
	} elsif ($cmd[0] eq 'TELNET_EC') {
		# backspace
		print "backspace\n";
		$heap->{input} =~ s/.$//;
		$input = '';
	}
	return;
  }


#  if ($input =~ tr[\x00-\x1F][]) {
#    personal_effect(
#      $session->ID,
#      "Try again without the control characters."
#    );
#    return;
#  }
#  print "got input [$input]\n";
my $o = ord($input);
print "input: [$input]=$o\n";

 if ($heap->{input_mode} ne 'line') {
	if ($heap->{telnet_echo} == 1 && $input ne "\n") {
		$kernel->yield(send_message => $input."\0");
	}
	if ($input =~ m/\x7f|\x08/) {
		if ($heap->{input}) { $heap->{input} =~ s/.$//; }
		$input = '';
	}
	if ($input eq "\n") {
		$kernel->yield(send_message => "");
		$input = delete $heap->{input};
		$heap->{input} = '';
	} else {
		$heap->{input} .= $input;
		return;
	}
  }

  print "input now: [$input]\n";
  
  ### those trixie hobbits
  $input =~ s/\0//g;
  
  ### Must authenticate before doing anything else.

  unless (defined $heap->{user}) {

    if ($input =~ /^\s*login\s+(\S+)\s+(\S.+?)\s*$/i) {
      my ($login, $pass) = ($1, $2);
      if (my $user = player_authenticate($login, $pass)) {
        personal_effect(
          $session->ID,
          "Welcome back, $login!"
        );
        player_appears($user, $session, "materializes.");
        $heap->{user} = $user;
        return;
      }

      personal_effect(
        $session->ID,
        "Invalid login or password, or you're already on!"
      );
      return;
    }

    if ($input =~ /^\s*signup\s+(\S+)\s+(\S.+?)\s*$/i) {
      my ($login, $pass) = ($1, $2);

      if (player_exists($login)) {
        personal_effect(
          $session->ID,
          "Sorry, '$login' is already in use.  Please try again."
        );
        return;
      }

      if (my $user = player_register($login, $pass)) {
        player_appears($user, $session, "materializes.");
        $heap->{user} = $user;
        return;
      }

      personal_effect(
        $session->ID,
        "Sorry.  Strange error trying to authenticate you.  Try again?",
        "Notify the owner of this multiwumpus if the problem persists."
      );
      return;
    }

    if ($input =~ /^\s*(instructions|info)\s*$/i) {
      personal_effect(
        $session->ID,
        'Multiwumpus $Revision$',
        "",
        "Based on WUMPUS 1 from the book BASIC COMPUTER GAMES, which is",
        "COPYRIGHT 1979  CREATIVE COMPUTING  MORRISTOWN, NJ.",
        "",
        "THE WUMPUS LIVES IN A CAVE OF 20 ROOMS.  EACH ROOM HAS 3",
        "TUNNELS LEADING TO OTHER ROOMS.  (LOOK AT A DODECAHEDRON TO",
        "SEE HOW THIS WORKS - IF YOU DON'T KNOW WHAT A DODECAHEDRON",
        "IS, ASK SOMEONE)",
        "",
        "HAZARDS:",
        "  BOTTOMLESS PITS:",
        "    TWO ROOMS HAVE BOTTOMLESS PITS IN THEM.  IF YOU GO THERE,",
        "    YOU FALL INTO THE PIT (& DIE!)",
        "  SUPERBATS:",
        "    TWO OTHER ROOMS HAVE SUPER BATS.  IF YOU GO THERE, A BAT",
        "    GRABS YOU AND TAKES YOU TO SOME OTHER ROOM AT RANDOM.",
        "  WUMPUS:",
        "    THE WUMPUS IS NOT BOTHERED BY OTHER HAZARDS (HE HAS SUCKER",
        "    FEET AND IS TOO BIG FOR A BAT TO LIFT).  USUALLY IT IS",
        "    ASLEEP, BUT IT WILL WAKE UP AND EAT YOU IF YOU STUMBLE",
        "    UPON IT.  IT MAY ALSO MOVE IF DISTURBED BY OTHER MEANS.",
        "",
        "YOU:",
        "  EACH TURN YOU MAY MOVE OR SHOOT A CROOKED ARROW",
        "  MOVING:",
        "    YOU CAN GO ONE ROOM (THROUGH ONE TUNNEL)",
        "  SHOOTING:",
        "    YOU HAVE UP TO FIVE ARROWS.  EACH ARROW CAN GO FROM ONE",
        "    TO FIVE ROOMS.  YOU MAY AIM IT TO AN ADJACENT ROOM, BUT",
        "    NOBODY KNOWS WHERE IT'LL GO AFTER THAT.",
        "  IF THE ARROW HITS THE WUMPUS, YOU WIN.",
        "  IF THE ARROW HITS YOU, YOU LOSE.",
        "",
        "WARNINGS:",
        "  WHEN YOU ARE ONE ROOM AWAY FROM WUMPUS OR HAZARD, THE",
        "  COMPUTER SAYS:",
        "    WUMPUS - 'YOU SMELL A WUMPUS!'",
        "    BAT    - 'BATS NEARBY!'",
        "    PIT    - 'YOU FEEL A DRAFT!'",
        "",
        "Good luck!",
      );
      return;
    }

    $kernel->yield("welcome_banner");
    return;
  }

  if ($input =~ /^\s*color\s+(on|off)\s*$/i) {
  	$heap->{colorize} = (lc($1) eq 'on') ? 1 : 0;
  	personal_effect($session->ID,"\0[bold;red]C\0[bold;green]o\0[bold;blue]l\0[bold;yellow]o\0[bold;white]r\0[reset] effects are now ".(($heap->{colorize} == 1) ? "\0[bold;green]On\0[reset]" : 'Off'));
  	return;
  }

  ### Quit is allowed from limbo.

  if ($input =~ /^\s*quit\s*(.*?)\s*$/i) {
    if ($heap->{is_in_limbo}) {
      $kernel->yield("disconnect");
      return;
    }
    player_quit($session, $1);
    return;
  }

  ### Guard condition.  Nothing else is a permitted from limbo.

  if ($heap->{is_in_limbo}) {
    $kernel->yield("remind_limbo");
    return;
  }

  if ($input =~ /\s*["']\s*(\S.+?)\s*$/) {
    player_speaks($session, $1);
    return;
  }

  if ($input =~ /^\s*(\d+)\s*$/) {
    player_moves($session, $1);
    return;
  }

  if ($input =~ /^\s*:\s*(\S.+?)\s*$/) {
    player_acts($session, $1);
    return;
  }

  if ($input =~ /^\s*f\s*(\d+)\s*$/i) {
    player_fire($session, $1);
    return;
  }

  if ($input =~ /^\s*l(ook)?\s*$/i) {
    player_look($session);
    return;
  }

  if ($input =~ /^\s*who\s*$/i) {
    player_list($session);
    return;
  }

  if ($input =~ /^\s*(\?|help)\s*$/i) {
    personal_effect(
      $session->ID,
      "| \0[bold;white]l\0[reset] or \0[bold;white]look\0[reset] to look around, but be careful!",
      "| Rooms are numbered.  Enter numbers to move around.",
      "| \0[bold;white]\"\0[reset]stuff to say stuff; \0[bold;white]:\0[reset]stuff to emote stuff.",
      "| \0[bold;white]quit\0[reset] (message) to quit, with some optional valediction.",
      "| \0[bold;white]f\0[reset]# to fire an arrow into a room.  Arrows travel.",
      "| \0[bold;white]about\0[reset] - Tells you about the game.",
      "| \0[bold;white]who\0[reset] - see who's online and how they're doing",
      "| \0[bold;white]color\0[reset] [on|off] - turn on/off ascii colors",
      "| Game rules are forthcoming.",
    );
    return;
  }

  if ($input =~ /^\s*about\s*$/i) {
    personal_effect(
      $session->ID,
      "| Hunt The Wumpus, originally by Gregory Yob.",
      "| - http://jerz.setonhill.edu/if/canon/Hunt_the_Wumpus.htm",
      "| - http://info.astrian.net/jargon/terms/w/wumpus.html",
      "| - http://en.wikipedia.org/wiki/Wumpus",
      "| Multiwumpus by Rocco Caputo <rcaputo\@pobox.com>",
      "| " . q($Id$),
      "| Uses POE for networking and multitasking.",
      "| - http://poe.perl.org/",
    );
    return;
  }

  personal_effect($session->ID, "Try \0[bold;white]?\0[reset] or help for help.");
}

# Send a message to the client.  Used by the game code to send back
# responses.

sub send_message {
  my ($heap, $message) = @_[HEAP, ARG0];
  return unless $heap->{client};
  return if $heap->{shutdown};
  # ARG1 is the line number from whitch the original sub was called, so
  # warnings can be issued on the offending line with the color booboo
  if ($heap->{colorize} == 1) {
    $message = colorize($message,$_[ARG1]);
  } else {
	$message =~ s/\0\[[^\]]+]//g;
  }
  $heap->{client}->put($message);
}

# Limbo state.  Enter it, disallow input for a short period of time,
# and then re-enter the game.

sub begin_limbo {
  my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];
  player_disappears($session, "fades away.");
  $heap->{is_in_limbo} = 1;
  $kernel->delay(cease_limbo => 10); # ten-second
  personal_effect(
    $_[SESSION]->ID,
    "---",
    "YOU'VE WON an all-expense-paid ten-second trip to beautiful",
    "Downtown LIMBO!  In just a few moments you'll be whisked back",
    "to the game.  You may also use the 'quit' command to leave if",
    "you're through hunting wumpuses or *koff* other players *koff*."
  );
}

sub cease_limbo {
  my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];
  $heap->{is_in_limbo} = 0;
  player_appears($heap->{user}, $session, "reappears!");
}

sub reminder_limbo {
  personal_effect(
    $_[SESSION]->ID,
    "Hey, you're dead!  Give it a rest for a few seconds."
  );
}

# Shutdown is less than absolutely trivial.  Clear a potential limbo
# timer, and shut down the socket.

sub do_shutdown {
  my $kernel = $_[KERNEL];
  $kernel->delay(cease_limbo => undef);
  $kernel->yield("shutdown");
}

#------------------------------------------------------------------------------
# MANAGE THE MULTIWUMPUS REALITY
#------------------------------------------------------------------------------

# TODO - This is a common pattern in certain classes of program.
# Consider making an abstract OO version of it available on the CPAN.

# NB - These numbers must never be greater than the total number of
# rooms.  It will be impossible to place hazards if they are, and the
# program will infinite-loop while looking for empty rooms.  In fact,
# playability demands they be rather small.
sub NUM_BATS     () { 2 }       # Number of bats.
sub NUM_PITS     () { 2 }       # Number of pits.
sub NUM_WUMP     () { 1 }       # Number of wumpi.

sub MAX_ARROWS   () { 5 }       # Quiver capacity.  Keep smallish.
sub USERS_FILE   () { "./users.db" }

# Player data structure.
sub PL_ID   () { 0 }
sub PL_LOC  () { 1 }
sub PL_NICK () { 2 }
sub PL_ARR  () { 3 }
sub PL_USER () { 4 }

# Room data structure.
sub RM_PLAYERS () { 0 }
sub RM_HAZARDS () { 1 }
sub RM_STUFF   () { 2 }

my %players;  # $players{$id} = $player;
my %rooms;    # $rooms{$room} = $room;
my %nicks;    # $nicks{$nick} = $player;

my %map;
my @rooms;

my ($users, %users);

### Map management section.

# Ensure a clean database closure, even in times of peril.

END {
  if ($users) {
    undef $users;
    untie %users;
  }
}

# Initialize the map.

sub map_setup {

  # Tie me hasheroo down, sport.
  $users = tie(
    %users, 'DB_File', USERS_FILE, O_RDWR | O_CREAT, 0666, $DB_HASH
  );
  die "can't tie " . USERS_FILE . ": $!" unless $users;

  # Reset map.
  %map = ( );
  foreach (
    [  1,  2,  5,  8 ],
    [  2,  1,  3, 10 ],
    [  3,  2,  4, 12 ],
    [  4,  3,  5, 14 ],
    [  5,  1,  4,  6 ],
    [  6,  5,  7, 15 ],
    [  7,  6,  8, 17 ],
    [  8,  1,  7,  9 ],
    [  9,  8, 10, 18 ],
    [ 10,  2,  9, 11 ],
    [ 11, 10, 12, 19 ],
    [ 12,  3, 11, 13 ],
    [ 13, 12, 14, 20 ],
    [ 14,  4, 13, 15 ],
    [ 15,  6, 14, 16 ],
    [ 16, 15, 17, 20 ],
    [ 17,  7, 16, 18 ],
    [ 18,  9, 17, 19 ],
    [ 19, 11, 18, 20 ],
    [ 20, 13, 16, 19 ],
  ) {
    my ($room, @exits) = @$_;
    foreach (@exits) {
      $map{$room}->{$_} = 1;
    }
  }

  @rooms = keys %map;

  # Clear hazards and room contents.
  foreach my $room (values %rooms) {
    $room->[RM_HAZARDS] = { };
    $room->[RM_STUFF] = {
      arrows => 0,
    };
  }

  # Random treasure.
  for (1..5) {
    my $room = $rooms[rand @rooms];
    $rooms{$room}->[RM_STUFF]->{arrows}++;
  }

  place_hazards(
    ("superbat") x NUM_BATS,
    ("pit")      x NUM_PITS,
    ("wumpus")   x NUM_WUMP,
  );
}

# Place several hazards randomly.

sub place_hazards {
  foreach my $hazard (@_) {
    my $loc = $rooms[rand @rooms];
    redo if (
      keys %{$rooms{$loc}->[RM_HAZARDS]} or
      keys %{$rooms{$loc}->[RM_PLAYERS]}
    );

    place_a_hazard($loc, $hazard);
  }
}

# Place a single hazard.  Notify people that things have changed.
# It's usually called during initial map setup, but it also is used to
# generate a new wumpus after an old one dies.

sub place_a_hazard {
  my ($loc, $hazard) = @_;

  $rooms{$loc}->[RM_HAZARDS]->{$hazard} = 1;

  if ($hazard eq "superbat") {
    room_effect(
      $loc,
      undef,
      "A giant bat flaps into the room and perches on a stalactite!"
    );
    area_effect(
      $loc,
      "You suddenly hear bats nearby!"
    );
    return;
  }

  if ($hazard eq "pit") {
    room_effect(
      $loc,
      undef,
      "There is a deafening roar, and a large part of the floor falls away!"
    );
    area_effect(
      $loc,
      "A deafening roar echoes through the cavern!"
    );
    return;
  }

  if ($hazard eq "wumpus") {
    room_effect(
      $loc,
      undef,
      (
        "A wumpus groggily wanders into the cavern " .
        "and finds a place to sleep.  " .
        "It doesn't seem to notice you."
      )
    );
    area_effect(
      $loc,
      "The stench of wumpus grows strong in your nostrils!"
    );
    return;
  }

  die "unknown hazard '$hazard'";
}

### Manage players.

# Find an existing player, verify their password, and return the data
# structure defining them.

sub player_authenticate {
  my ($login, $password) = @_;

  return unless player_exists($login);
  return if exists $nicks{$login};
  my $user = thaw($users{$login});

  my $salt = substr($user->{pass}, 0, 2);
  my $test_password = crypt($password, $salt);
  return unless $user->{pass} eq $test_password;

  return $user;
}

# A simple accessor.

sub player_exists {
  my $login = shift;
  return exists $users{$login};
}

# Create a new player if the nickname doesn't already exist.

sub player_register {
  my ($login, $password) = @_;

  return if player_exists($login);
  my $user = { };

  my @salt = ("a".."z","A".."Z",0..9);
  my $salt = $salt[rand @salt] . $salt[rand @salt];
  $user->{pass} = crypt($password, $salt);
  $user->{nick} = $login;

  $user->{plays} = 0;
  $user->{quits} = 0;
  $user->{shot}  = 0;
  $user->{eaten} = 0;
  $user->{fell}  = 0;
  $user->{fired} = 0;
  $user->{kills} = 0;
  $user->{wumps} = 0;

  $users{$login} = freeze($user);

  return $user;
}

# A new player appears in the world.  Set up their initial inventory,
# find a safe place for them, and put them down.  Let others know
# they've arrived.

sub player_appears {
  my ($user, $client, $message) = @_;
  my $client_id = $client->ID;

  my $location;
  while (1) {
    $location = $rooms[rand @rooms];
    last unless keys %{$rooms{$location}->[RM_HAZARDS]};
  }

  $user->{plays}++;
  my $nick = $user->{nick};

  my $player = [
    $client_id,     # PL_ID
    $location,      # PL_LOC
    $nick,          # PL_NICK
    MAX_ARROWS,     # PL_ARR
    $user,          # PL_USER
  ];

  $players{$client_id}                          = $player;
  $rooms{$location}->[RM_PLAYERS]->{$client_id} = $player;
  $nicks{$nick}                                 = $player;

  effect(
    $client_id,
    undef,
    "$nick $message.",
    "You hear a soft *bamf* nearby.",
    undef
  );
  describe($client_id, $location);
}

# Remove a player from the world.

sub player_disappears {
  my ($client, $message) = @_;
  my $client_id = $client->ID;

  my $player = $players{$client_id};

  effect(
    $client_id,
    undef,
    "$players{$client_id}->[PL_NICK] $message.",
    "You hear a soft *moof* nearby.",
    undef
  );

  # Record the user on file.
  $users{$player->[PL_NICK]} = freeze($player->[PL_USER]);

  delete $players{$client_id};
  delete $rooms{$player->[PL_LOC]}->[RM_PLAYERS]->{$client_id};
  delete $nicks{$player->[PL_NICK]};
}

# Register that a player has quit, possibly also telling nearby
# people.

sub player_quit {
  my ($client, $message) = @_;
  my $client_id = $client->ID;

  my $quit = "$players{$client_id}->[PL_NICK] has quit";
  if (defined $message and length $message) {
    $quit .= ": $message";
  }
  else {
    $quit .= ".";
  }
  effect(
    $client_id,
    "Bye!",
    $quit,
    undef,
    undef
  );
  $players{$client_id}->[PL_USER]->{quits}++;

  $poe_kernel->post($client_id, "disconnect");
}

# Spoken messages are heard in-room, and talking is heard nearby.

sub player_speaks {
  my ($client, $message) = @_;
  my $client_id = $client->ID;
  my $pl_loc = $players{$client_id}->[PL_LOC];

  effect(
    $client_id,
    "You say: $message",
    "$players{$client_id}->[PL_NICK] says: $message",
    "You hear talking nearby.",
    undef
  );
}

# Like speaking, except the player acts something out.  In MUD terms,
# this is an emote.

sub player_acts {
  my ($client, $action) = @_;
  my $client_id = $client->ID;
  effect(
    $client_id,
    "$players{$client_id}->[PL_NICK] $action",
    "$players{$client_id}->[PL_NICK] $action",
    "You hear someone moving about nearby.",
    undef,
  );
}

# The player is attempting to fire an arrow.  Handle all that.
#
# TODO - This function is perhaps way too long.

sub player_fire {
  my ($client, $room) = @_;
  my $client_id = $client->ID;
  my $nick = $players{$client_id}->[PL_NICK];
  my $origin = $players{$client_id}->[PL_LOC];

  unless ($players{$client_id}->[PL_ARR]) {
    effect(
      $client_id,
      "You have no arrows left!",
      "$nick looks sadly at their bow.",
      undef,
      undef
    );
    return;
  }

  unless (exists $map{$origin}->{$room}) {
    effect(
      $client_id,
      (
        "There's no path to $room.  Exits: " .
        join(" ", sort {$a<=>$b} keys %{$map{$origin}})
      ),
      "$nick aims their bow around nervously.",
      undef
    );
    return;
  }

  effect(
    $client_id,
    "You let an arrow fly!",
    "$nick fires into cavern $room.",
    "You hear a nearby twang!",
    undef
  );

  $players{$client_id}->[PL_USER]->{fired}++;
  $players{$client_id}->[PL_ARR]--;
  my $arrow_moves = int(rand(5)+1);
  my $arrow_room = $room;

  while ($arrow_moves--) {
    my $new_room;
    my @exits = keys %{$map{$arrow_room}};

    my $bailout = 100;
    while ($bailout--) {
      $new_room = $exits[rand @exits];
      last unless $new_room == $arrow_room;
    }
    $arrow_room = $new_room;

    my @targets = (
      keys %{$rooms{$arrow_room}->[RM_PLAYERS]},
      keys %{$rooms{$arrow_room}->[RM_HAZARDS]}
    );

    # You missed!  Note: Crowded rooms are easier to hit things in.
    unless (rand() < @targets / 3) {
      room_effect(
        $arrow_room,
        undef,
        "An arrow whizzes past you!"
      );

      # If the wumpus was here, we should maybe move it?
      if (exists $rooms{$arrow_room}->[RM_HAZARDS]->{wumpus}) {
        # 50% chance of movement
        if (rand() < 0.5) {
          my @wumpus_exits = keys %{$map{$arrow_room}};
          my $wumpus_move = $wumpus_exits[rand @wumpus_exits];

          # TODO - Move the wumpus.  This requires a more generic
          # item/creature framework if it will be done properly..
        }
      }
      next;
    }

    my $target = $targets[rand @targets];

    if ($target =~ /^\d+$/) {
      my $target_nick = $players{$target}->[PL_NICK];
      my $target_arr  = $players{$target}->[PL_ARR];

      my @room = (
        "Thunk!  $target_nick is hit in the back with an arrow!"
      );

      if ($target_arr > 1) {
        push @room, "You notice they have some arrows!";
      }
      elsif ($target_arr == 1) {
        push @room, "Hey, they've got an arrow.  Maybe you can use it.";
      }

      effect(
        $target,
        "Whizzzzzthunk!  You've been shot!  -More-  You die!",
        \@room,
        "Serious urking and thudding from an adjoining cavern!",
        undef
      );
      $players{$target}->[PL_USER]->{shot}++;
      $players{$client_id}->[PL_USER]->{kills}++;
      $rooms{$arrow_room}->[RM_STUFF]->{arrows} += $target_arr;
      $poe_kernel->post($target, "begin_limbo");
      return;
    }

    if ($target eq "pit") {
      room_effect(
        $arrow_room,
        undef,
        "An arrow whizzes into the room and clatters into a pit!"
      );
      area_effect(
        $arrow_room,
        "You hear faint clattering from $arrow_room."
      );
      return;
    }

    if ($target eq "superbat") {
      # TODO - Possibly carrying you with it.
      room_effect(
        $arrow_room,
        undef,
        "An arrow hits the large bat, sending it flying away!"
      );
      area_effect(
        $arrow_room,
        (
          "An almost impossibly loud squeal fades away from $arrow_room."
        )
      );

      my $bat_drop;
      my $bailout = 100;
      while ($bailout--) {
        $bat_drop = $rooms[rand @rooms];
        last unless keys %{$rooms{$bat_drop}->[RM_HAZARDS]};
      }
      delete $rooms{$arrow_room}->[RM_HAZARDS]->{"superbat"};

      place_a_hazard($bat_drop, "superbat");
      return;
    }

    if ($target eq "wumpus") {
      # TODO - Randomise map?

      effect(
        $client_id,
        "Congratulations!  You got the wumpus!  Generating a new one...",
        "An arrow shoots in from somewhere, striking the wumpus dead!",
        "A terrible roar erupts from somewhere close by, then dies away!",
        "Congratulations!  $nick got the wumpus!  Generating a new one..."
      );

      $players{$client_id}->[PL_USER]->{wumps}++;
      delete $rooms{$arrow_room}->[RM_HAZARDS]->{wumpus};

      place_hazards("wumpus");

      foreach (values %players) {
        if ($_->[PL_ARR] < MAX_ARROWS) {
          personal_effect($_->[PL_ID], "Your quiver refills with arrows!");
          $_->[PL_ARR] = MAX_ARROWS;
        }
      }

      return;
    }

    warn "unknown target '$target'";
  }

  room_effect(
    $arrow_room,
    undef,
    "An arrow lands in the dirt."
  );

  $rooms{$arrow_room}->[RM_STUFF]->{arrows}++;
}

# Respond to the user with a list of players and their statistics.

sub player_list {
  my $client_id = shift->ID;
  my @nicks = sort keys %nicks;

  my $data_format =
    "%-20.20s %5d %5d %5d %5d %5d %5d %5d %7.2f %5d";
  my $head_format =
    "%-20.20s %5.5s %5.5s %5.5s %5.5s %5.5s %5.5s %5.5s %7.7s %5.5s";

  personal_effect(
    $client_id,
    "--- People Online ---",
    sprintf(
      $head_format,
      "Nickname",
      "Plays", "Quits",
      "Shot", "Eaten", "Fell",
      "Shots", "Hits",
      "Acc%",
      "Wins",
    )
  );

  foreach my $nickname (sort keys %nicks) {
    my $user = $nicks{$nickname}->[PL_USER];
    personal_effect(
      $client_id,
      sprintf(
        $data_format,
        $nickname,
        $user->{plays}, $user->{quits},
        $user->{shot}, $user->{eaten}, $user->{fell},
        $user->{fired}, $user->{kills},
        ($user->{kills} / ($user->{fired} || 1)) * 100,
        $user->{wumps},
      )
    );
  }

  personal_effect(
    $client_id,
    sprintf(
      $head_format,
      ( "-------------------------------------------------------" ) x 20
    )
  );
}

# The player is looking around the room.  This triggers describe.

sub player_look {
  my $client_id = shift->ID;
  my $room = $players{$client_id}->[PL_LOC];
  describe($client_id, $room);
}

# Describe the room to a player.  This triggers room-entry side
# effects.
#
# TODO - It may be better to separate room-entry side effects from the
# room description.

sub describe {
  my ($client_id, $room) = @_;

  my @hazards = keys %{$rooms{$room}->[RM_HAZARDS]};

  my $desc = "You are in $room.";
  $desc .= "  The coast appears clear.  For now." unless @hazards;
  personal_effect(
    $client_id,
    "---",
    $desc,
  );

  my $nick = $players{$client_id}->[PL_NICK];

  if (@hazards) {

    my $hazard = $hazards[0];
    if ($hazard eq "wumpus") {
      effect(
        $client_id,
        "Oops!  You bumped a wumpus!  -More-  You die!",
        "Horrors!  $nick has been eaten by a wumpus, arrows and all!",
        "There's a screaming and a threashing near you, then silence.",
        undef
      );
      $players{$client_id}->[PL_USER]->{eaten}++;
      $poe_kernel->post($client_id, "begin_limbo");
      return;
    }

    if ($hazard eq "superbat") {
      effect(
        $client_id,
        "-More-  ZAP!  Super bat snatch!  Elsewhereville for you!",
        "A giant bat whisks $nick away with a flapping and a thrashing!",
        (
          "Flapping, thrashing, and shouting erupt from an adjacent cavern " .
          "then fade into the distance."
        ),
        undef
      );

      # Find a room without a hazard.
      my $bat_drop;
      while (1) {
        $bat_drop = $rooms[rand @rooms];
        last unless keys %{$rooms{$bat_drop}->[RM_HAZARDS]};
      }
      delete $rooms{$room}->[RM_HAZARDS]->{"superbat"};

      $rooms{$bat_drop}->[RM_PLAYERS]->{$client_id} =
        delete $rooms{$room}->[RM_PLAYERS]->{$client_id};
      $players{$client_id}->[PL_LOC] = $bat_drop;
      describe($client_id, $bat_drop);

      $rooms{$bat_drop}->[RM_HAZARDS]->{"superbat"} = 1;

      effect(
        $client_id,
        "Your bat finds a nice stalactite and makes itself comfortable.",
        [
          "$nick is carried into the room by a huge bat!",
          "The bat finds a nice stalactite and makes itself comfortable."
        ],
        "You hear flapping approach, and an undignified wet-sack sort of thud.",
        undef
      );
      return;
    }

    if ($hazard eq "pit") {
      effect(
        $client_id,
        "Yyyiiiieeee!  You fell in a pit!  -More-  You die!",
        "You hear a scream and turn to see $nick drop out of sight!",
        "Someone yells nearby!  Their screams rapidly fade into infinity.",
        undef
      );
      $players{$client_id}->[PL_USER]->{fell}++;
      $poe_kernel->post($client_id, "begin_limbo");
      return;
    }

    warn "unknown hazard '$hazard'";
  }

  my @sounds;
  foreach my $neighbor (keys %{$map{$room}}) {
    foreach my $hazard (keys %{$rooms{$neighbor}->[RM_HAZARDS]}) {
      if ($hazard eq "pit") {
        push @sounds, "You feel a draft!";
      }
      elsif ($hazard eq "superbat") {
        push @sounds, "Bats nearby!";
      }
      elsif ($hazard eq "wumpus") {
        push @sounds, "You smell a wumpus!";
      }
    }
  }
  if (@sounds) {
    personal_effect($client_id, ">>> HEY!  @sounds");
  }

  personal_effect(
    $client_id,
    "Paths lead off to: " . join(" ", sort {$a<=>$b} keys %{$map{$room}})
  );
  my @people = grep { $_ != $client_id } keys %{$rooms{$room}->[RM_PLAYERS]};
  if (@people) {
    personal_effect(
      $client_id,
      (
        "There are people here: " .
        join(
          "; ",
          map { $players{$_}->[PL_NICK] } @people
        )
      )
    );
  }
  else {
    personal_effect($client_id, "You are all alone.");
  }

  my $stuff = $rooms{$players{$client_id}->[PL_LOC]}->[RM_STUFF];
  if ($stuff->{arrows}) {
    my $message = "You come across $stuff->{arrows} arrow";
    $message .= "s" if $stuff->{arrows} > 1;
    $message .= ".";
    if ($players{$client_id}->[PL_ARR] < MAX_ARROWS) {
      my $take = MAX_ARROWS - $players{$client_id}->[PL_ARR];
      if ($stuff->{arrows} <= $take) {
        $take = $stuff->{arrows};
        if ($take > 1) {
          $message .= "  You take them.";
        }
        else {
          $message .= "  You take it.";
        }
      }
      else {
        $message .= "  You take $take.";
      }

      room_effect(
        $room,
        $client_id,
        (
          "$nick picks up " .
          ( ($take > 1)
            ? "some arrows."
            : "an arrow."
          )
        )
      );

      my $new_count = $players{$client_id}->[PL_ARR] += $take;
      $stuff->{arrows} -= $take;

      $message .= "  Now you have $new_count.";
    }
    else {
      $message .= "  You can't carry any more, however.";
    }
    personal_effect($client_id, $message);
  }
}

# The player attempts to move.

sub player_moves {
  my ($client, $new_room) = @_;
  my $client_id = $client->ID;

  my $nick = $players{$client_id}->[PL_NICK];
  my $old_room = $players{$client_id}->[PL_LOC];

  if (exists $map{$old_room}->{$new_room}) {
    room_effect(
      $old_room,
      $client_id,
      "$nick leaves towards $new_room."
    );

    $rooms{$new_room}->[RM_PLAYERS]->{$client_id} =
      delete $rooms{$old_room}->[RM_PLAYERS]->{$client_id};
    $players{$client_id}->[PL_LOC] = $new_room;

    room_effect(
      $new_room,
      $client_id,
      "$nick arrives from $old_room."
    );
    area_effect(
      $new_room,
      "You hear movement in a nearby cavern."
    );

    describe($client_id, $new_room);
    return;
  }
  else {
    effect(
      $client_id,
      (
        "There's no path to $new_room.  Exits: " .
        join(" ", sort {$a<=>$b} keys %{$map{$old_room}})
      ),
      "$nick confusedly scans the cavern.",
      undef,
      undef
    );
    return;
  }
}

#------------------------------------------------------------------------------
# MANAGE GAMESPACE VISUAL AND AUDIO EFFECTS
#------------------------------------------------------------------------------

# Send a private message to one player.

sub personal_effect {
  my $client_id = shift;
  my @messages = map { (ref eq "ARRAY") ? @$_ : $_ } @_;

  foreach my $message (@messages) {
  # see note in send_message
    $poe_kernel->post($client_id, send_message => $message, (caller)[2]);
  }
}


sub colorize {
  my ($msg, $line) = @_;

  my $color = sub {
  	my $c = $_[0];
  	
  	# should we support number colors?
  	if ($c =~ m/^\d+$/ || $c =~ m/^\d+;\d+$/) {
  		return "\e[${c}m";
  	} else {
  		if ($c =~ m/^(.+);(.*)/) {
  			unless (exists($colors{$1})) {
  				warn "undef color '$1' at line $line\n";
  				return "";
  			}
  			unless (exists($colors{$2})) {
  				warn "undef color '$2' at line $line\n";
  				return "";
  			}
  			return "\e[$colors{$1};$colors{$2}m";
  		} else {
  			unless (exists($colors{$c})) {
  				warn "undef color '$c' at line $line\n";
  				return "";
  			}
  			return "\e[$colors{$c}m";
  		}
  	}
  };
  
  $msg =~ s/\0\[([^\]]+)\]/$color->($1)/eg;

  return $msg;
}

# Send a public effect to everyone in a room, possibly excepting a
# single player.  That player is usually the source of the effect.

sub room_effect {
  my ($room, $cause, @messages) = @_;

  foreach my $player (values %{$rooms{$room}->[RM_PLAYERS]}) {
    next if $cause and $cause == $player->[PL_ID];
    personal_effect($player->[PL_ID], @messages);
  }
}

# Send a public effect to everyone in surrounding rooms.  This is
# usually a muted version of the room effect.  If someone speaks, for
# example, you hear their words in-room, but players in connected
# rooms only know that someone is talking nearby.

sub area_effect {
  my ($room, @messages) = @_;

  foreach my $neighbor (keys %{$map{$room}}) {
    room_effect($neighbor, undef, @messages);
  }
}

# Broadcast a message to everyone who is not the player.  This is to
# be used exceedingly sparingly.  Overuse blurs the distinction
# between rooms, making the gamespace feel vague and indistinct.

sub global_effect {
  my ($client_id, $message) = @_;

  foreach my $player (values %players) {
    next if $player->[PL_ID] == $client_id;
    personal_effect($player->[PL_ID], $message);
  }
}

# Quite often something will generate personal, room-except-person,
# and area effects.  This wrapper provides a single function call that
# translates into all three, plus a potential global effect.

sub effect {
  my ($client_id, $personal, $room, $area, $global) = @_;
  my $place = $players{$client_id}->[PL_LOC];

  personal_effect($client_id, $personal) if defined $personal;
  room_effect($place, $client_id, $room) if defined $room;
  area_effect($place, $area)             if defined $area;
  global_effect($client_id, $global)     if defined $global;
}

__END__

- Items
  + Corpses (with arrows)
    Sort of done now, but without the corpse.  That is, dying players
    will leave their arrows behind if it's not a pit room.
  - Graffitti

- Monsters
  - Wumpus should move periodically.
  - Bats should also occasionally move around.
  - Bats should grab someone when they arrive in a room.
  - Hazard effects should not happen during describe.
  - Pits should occasionally rearrange themselves.

- Environmental effects.
  - Very occasional breezes.
  - Very occasional water dripping and stuff.
    Maybe implement them as random input timeouts.  If someone stays
    still long enough, they start to relax and notice the scenery.

