#!/usr/bin/perl

use strict;
use lib qw(/projects/lib);
use POE qw(Component::Server::Talker);

POE::Session->create(
	inline_states => {
		_start => sub {
			my ($kernel, $heap) = @_[KERNEL, HEAP];
			
			POE::Component::Server::Talker->spawn(
				Alias           => 'talker',					# talker is default
				ListenPort      => 9995,							# port to listen on
				Greeting		=> 'return "Welcome to POE Talker v'.
					$POE::Component::Server::Talker::VERSION.
					' ($^O)";',									# shown on connection
				DataDir			=> '/projects/talker/data',		# no trailing slash
				MOTD			=> 'motd',						# no beginning slash
				UsersFile		=> 'users.db',					# no beginning slash
				AllowGuests		=> 1,							# 0 = off, 1 = on
				LogLevel		=> 4,							# 4=debug, 3=less info, 2=quiet, 1=really quiet
				LoginTimeOut	=> 30,							# Time to login
				ModulesDir		=> '/projects/talker/modules',	# relative location of modules
				# Not implemented, yet
				# MUST use a client that supports this! -Make your own :)
				#SslType			=> 2,							# 1 = implicit, 2 '.secure ssl' command
				#SslPrivateKey	=> 'data/plain-rsa.pem',
				#SslCert			=> 'data/plain-cert.pem',
			);
			
			$kernel->call(talker => 'register');
		},
		_stop => sub {
			my ($kernel, $heap) = @_[KERNEL, HEAP];

		},
		_default => sub {
			my ($kernel, $heap, $event, $arg) = @_[KERNEL, HEAP, ARG0, ARG1];
			
			return 0 if ($event =~ m/^_/);
			
			if (ref($arg->[0]) eq 'HASH') {
				print "unhandled event $event\n";
				#require Data::Dumper;
				#delete $arg->[0]->{con_session};
				#print Data::Dumper->Dump([$arg])."\n";
			} else {
				print "unhandled event $event\n";
			}
			
			return 0;
		},
		talker_registered => sub {
			my ($kernel, $heap) = @_[KERNEL, HEAP];
			
			# put start code here, or use a call in _start like this: $kernel->call(talker => 'register');
			# before using talker
		},
		talker_accept => sub {
			my ($kernel, $heap, $data) = @_[KERNEL, HEAP, ARG0];

			# allow this ip address?
#			if ($data->{peer_addr} eq '10.0.2.1') {
#				print "connection denied from $data->{peer_addr}:$data->{peer_port} on $data->{local_ip}:$data->{local_port}\n";
#				# disallow
#				return 0;
#			} else {
				print "connection accepted from $data->{peer_addr}:$data->{peer_port} on $data->{local_ip}:$data->{local_port}\n";
				# allow
				return 1;
#			}
		},
		talker_write_log => sub {
			my ($kernel, $heap, $o) = @_[KERNEL, HEAP, ARG0];
			
			my $datetime = localtime();
			print "[$datetime][$o->{type}$o->{sender}] $o->{msg}\n";
		},
	},
);

$poe_kernel->run();
